
using System.Collections.Generic;
using Jypeli;
namespace matopeli;

/// @author Paulus Kopperoinen
/// @version 30.10.2023
/// <summary>
/// Perinteinen matopeli, jossa tavoitteena kasvattaa matoa mahdollisimman suureksi omenia syömällä.
/// </summary>

public class Matopeli : PhysicsGame
{
    private static readonly string[] Tasonkuva = {
        "#####################",
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#         m         #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #", 
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#o                  #",
        "#####################",
    };
    
    
    private Objektit _mato;
    private Objektit _omena;
    private Objektit _matoMuu;

    private IntMeter Pistelaskuri;
    public IntMeter Ennatyslaskuri;

    private readonly List<Vector> MatoSijainti = new List<Vector>();
    private readonly List<PhysicsObject> MatoKeho = new List<PhysicsObject>();

    private int Leveys;
    private int Korkeus;
    private double Nopeus;

    private Vector _omenaAlkuPaikka;
    
    private static readonly Image Seinakuva = LoadImage("seina.png");
    private static readonly Image Matopaakuva = LoadImage("matopaa.png");
    private readonly Image _matokehokuva = LoadImage("matokeho.png");
    private readonly Image _omenakuva = LoadImage("omena.png");
    
    private bool _suuntaYlos;
    private bool _suuntaAlas;
    private bool _suuntaOikea;
    private bool _suuntaVasen;

    private bool _matoLiikkunut;
    private bool _omenaKentalla;
    
    /// <summary>
    /// Kun peli käynnistetään ensimmäisen kerran
    /// </summary>
    public override void Begin()
    {
        AloitaUusiPeli();
    }

    
    /// <summary>
    /// Poistaa vanhan pelin tiedot ja alustaa uuden pelin
    /// </summary>
    private void AloitaUusiPeli()
    {
        MatoKeho.Clear();
        MatoSijainti.Clear();
        ClearAll();
        Asetukset();
        TileMap tiles = TileMap.FromStringArray(Tasonkuva);
        tiles['#'] = LuoSeina;
        tiles['m'] = LuoMato;
        tiles.Insert(Leveys,Korkeus); 
        _omenaAlkuPaikka = _mato.Position;
        Asetaohjaimet();
        LuoAikalaskuri();
        LuoPistelaskuri();
        LuoOmena();
        Level.Background.Color = Color.LightGreen;
    }
    
    
    /// <summary>
    /// Määrittää ohjaimet, joita pelaaja voi käyttää ja mitä niitä painamalla tapahtuu
    /// </summary>
    private void Asetaohjaimet()
    {
        Keyboard.Listen(Key.Up, ButtonState.Pressed, VaihdaSuuntaYlos, "Turn upwards");
        Keyboard.Listen(Key.Down, ButtonState.Pressed, VaihdaSuuntaAlas, "Turn downwards");
        Keyboard.Listen(Key.Left, ButtonState.Pressed, VaihdaSuuntaVasen , "Turn left");
        Keyboard.Listen(Key.Right, ButtonState.Pressed, VaihdaSuuntaOikea, "Turn right");
        Keyboard.Listen(Key.Enter, ButtonState.Pressed, LiikutaMatoa, "New Game");
    }


    /// <summary>
    /// Laskee pisteet, joita kerätään omenoita syömällä
    /// </summary>
    void LuoPistelaskuri()
    {
        Pistelaskuri = new IntMeter(0);               
        Label pistenaytto = new Label(); 
        pistenaytto.X = Screen.Left + 500;
        pistenaytto.Y = Screen.Top - 5;
        pistenaytto.TextColor = Color.Brown;
        pistenaytto.Color = Color.LightGreen;
        pistenaytto.Title = "Pisteet: ";
        pistenaytto.BindTo(Pistelaskuri);
        Add(pistenaytto);
    }
    

    /// <summary>
    /// Pelin asetukset
    /// </summary>
    private void Asetukset()
    {
    Leveys = 40;
    Korkeus = 40;
    _suuntaYlos = false;
    _suuntaAlas = false;
    _suuntaOikea = true; 
    _suuntaVasen = false;
    Nopeus = 0.1;
    _matoLiikkunut = false;
    _omenaKentalla = true;
    }
    
    
    /// <summary>
    /// Päättää pelin ja aloitaa uuden, kun mato kuolee
    /// </summary>
    private void MatoKuolee()
    {
        AloitaUusiPeli();
    }
    
    /// <summary>
    /// Luo madon kentälle
    /// </summary>
    /// <returns></returns>
    private PhysicsObject LuoMato()
    {
        _mato = new Objektit(Leveys-5, Korkeus-5);
        _mato.Image = Matopaakuva;
        Add(_mato);
        _mato.Tag = "mato";
        
        AddCollisionHandler(_mato, "omena", delegate(IPhysicsObject _, IPhysicsObject _) {MatoSyoOmenan();  });
        AddCollisionHandler(_mato, "seina", delegate(IPhysicsObject _, IPhysicsObject _) {MatoKuolee();  });
        AddCollisionHandler(_mato, "matokeho", delegate(IPhysicsObject _, IPhysicsObject _) {MatoKuolee();  });
        
        return _mato;
    }

    
    /// <summary>
    /// Luodaan omena johonkin satunnaiseen kohtaan kartalla, joka ei ole samassa kohtaa madon kanssa.
    /// </summary>
    private void LuoOmena()
    {
        Vector paikka = ArvoOmenaPaikka();
        _omena = new Objektit(Korkeus-10, Leveys-10);
        _omena.Shape = Shape.Circle;
        _omena.Position = paikka;
        _omena.Image = _omenakuva;
        _omena.Tag = "omena";
        Add(_omena);
        _omenaKentalla = true;
    }

    
    /// <summary>
    /// tehtävä on arpoa omenalle paikka, johon se luodaan. En ole vieläkään varma kuinka hyvin tämä toimii.
    /// </summary>
    private Vector ArvoOmenaPaikka()
    {
        Vector omenaPaikka;
        double x;
        double y;

        y = RandomGen.NextInt(-10, 5);
        y = _omenaAlkuPaikka.Y + y * Leveys;
            
        x = RandomGen.NextInt(-8, 8);
        x = _omenaAlkuPaikka.X + x * Leveys;

        if (x - _mato.X < Leveys && y - _mato.Y < Leveys)
        {
            y = RandomGen.NextInt(-10, 5);
            y = _omenaAlkuPaikka.Y + y * Leveys;
            
            x = RandomGen.NextInt(-8, 8);
            x = _omenaAlkuPaikka.X + x * Leveys;
        }
        
        omenaPaikka = new Vector(x, y);
        return omenaPaikka;
    }
        
    
    /// <summary>
    /// Luo kentän reunat
    /// </summary>
    private PhysicsObject LuoSeina()
    {
        PhysicsObject seina = new PhysicsObject(Leveys, Korkeus);
        seina.MakeStatic();
        seina.Color = Color.Green;
        seina.Image = Seinakuva;
        seina.Tag = "seina";
        return seina;
    }
    
    
    /// <summary>
    /// Tehdään aliohjelma, joka luo uusia osia madon perään.
    /// </summary>
    public PhysicsObject LuoMatoKeho()
    {
        _matoMuu = new Objektit(Leveys-6, Korkeus-6);
        _matoMuu.Tag = "matokeho";
        _matoMuu.Image = _matokehokuva;
        Vector paikka = MatoSijainti[MatoSijainti.Count -1 - MatoKeho.Count];
        _matoMuu.Position = paikka;
        Add(_matoMuu);
        MatoKeho.Add(_matoMuu);
        AddCollisionHandler(_matoMuu, "omena", delegate(IPhysicsObject _, IPhysicsObject _) {LuoUusiOmena();  });
        return _matoMuu;
    }

    
    /// <summary>
    /// Jos omena luodaan madon kehon sisään, tämä aliohjelma luo uuden omenan vanhan tilalle
    /// </summary>
    private void LuoUusiOmena()
    {
        _omena.Destroy();
        _omenaKentalla = false;
    }
    
    
    /// <summary>
    /// Tekee tarvittavat toimet, kun mato onnistuu syömään omenan
    /// </summary>
    private void MatoSyoOmenan()
    {
        _omena.Destroy();
        Pistelaskuri.Value += 1;
        _omenaKentalla = false;
    }
    
    
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä ylös.
    /// </summary>
     private void VaihdaSuuntaYlos()
        {
            
            if (_suuntaAlas)
            {
                _suuntaOikea = false;
                _suuntaVasen = false;
                _suuntaYlos = false;
                _suuntaAlas = true;
            }
            else
            {
                _suuntaOikea = false;
                _suuntaVasen = false;
                _suuntaYlos = true;
                _suuntaAlas = false;
            }
        }
     
    
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä alas.
    /// </summary>
     private void VaihdaSuuntaAlas()
        {
            if (_suuntaYlos)
            {
                _suuntaOikea = false;
                _suuntaVasen = false;
                _suuntaYlos = true;
                _suuntaAlas = false;
            }
            else
            {
                _suuntaOikea = false;
                _suuntaVasen = false;
                _suuntaYlos = false;
                _suuntaAlas = true;
            }
        }
      
     
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä oikeaan.
    /// </summary>
     private void VaihdaSuuntaOikea()
        {
            if (_suuntaVasen)
            {
                
                _suuntaOikea = false;
                _suuntaVasen = true;
                _suuntaYlos = false;
                _suuntaAlas = false;
            }
            else
            {
                _suuntaOikea = true;
                _suuntaVasen = false;
                _suuntaYlos = false;
                _suuntaAlas = false;
            }

        }
       
     
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä vasempaan.
    /// </summary>
     private void VaihdaSuuntaVasen()
        {
            if (_suuntaOikea)
            {
                _suuntaOikea = true;
                _suuntaVasen = false;
                _suuntaYlos = false;
                _suuntaAlas = false;
            }
            else
            {
                _suuntaOikea = false;
                _suuntaVasen = true;
                _suuntaYlos = false;
                _suuntaAlas = false;
            }
        }
     
     
     /// <summary>
     /// Aikalaskuri toimii pelin moottorina ja antaa PaivitaKentälle tiedon, kun pitää tarkistaa mitä kentällä on tapahtunut.
     /// </summary>
     private void LuoAikalaskuri()
        {
            Timer aikalaskuri = new Timer();
            aikalaskuri.Interval=Nopeus;
            aikalaskuri.Start();
            aikalaskuri.Timeout += PaivitaKentta; 
        }

     
     /// <summary>
     /// Kun ajastimesta loppuu aika, siirtyy peli tähän funktioon, jonka tarkoituksena on hahmottaa, mitä kaikkea pelissä on tapahtunut ja miten siihen tulee reagoida.
     /// </summary>
     private void PaivitaKentta()
     {
         if (_matoLiikkunut == false)
         {
             return;
         }
         
         if (_omenaKentalla == false)
         {
             LuoMatoKeho();
             LuoOmena();
         }
         
         MatoSijainti.Add(new Vector(_mato.X,_mato.Y));
         LiikutaMatoa();
         LiikutaKehoa();
         MatoSijainti.Add(new Vector(_mato.X,_mato.Y));  
     }
     
     
     /// <summary>
     /// LiikutaMatoa aliohjelman tehtävä on selvittää tieto siitä, mihin suuntaan on tarkoitus liikkua ja liikuttaa matoa siihen suuntaan.
     /// </summary>
    private void LiikutaMatoa()
    {
        _matoLiikkunut = false;
        if (_suuntaYlos)
        {
            _mato.Y = _mato.Y + Leveys;
            _mato.Angle = Angle.FromDegrees(90);
        }
        if (_suuntaAlas)
        {
            _mato.Y= _mato.Y - Leveys;
            _mato.Angle = Angle.FromDegrees(270);
        }
        if (_suuntaOikea)
        {
            _mato.X= _mato.X + Leveys;
            _mato.Angle = Angle.FromDegrees(0);
           
        }
        if (_suuntaVasen)
        {
            _mato.X= _mato.X - Leveys;
            _mato.Angle = Angle.FromDegrees(180);
            
        }
        _matoLiikkunut = true;
    }

    
    /// <summary>
    /// Liikutakehoa, aliohjelman tehtävänä on liikuttaa madon kehoa sen pään mukaisesti.
    /// </summary>
    private void LiikutaKehoa()
    {
        if (MatoKeho.Count==0)
        {
            return;
        }
        _matoLiikkunut = false;
        MatoKeho[0].X = MatoSijainti[MatoSijainti.Count - 1].X;
        MatoKeho[0].Y = MatoSijainti[MatoSijainti.Count - 1].Y;
        
        for (int i = 1; i < MatoKeho.Count; i++)
        {
            MatoKeho[i].X = MatoSijainti[MatoSijainti.Count - 2*i].X;
            MatoKeho[i].Y = MatoSijainti[MatoSijainti.Count - 2*i].Y;
        }

        _matoLiikkunut = true;
    }


    public class Objektit : PhysicsObject
    {
         public Objektit(double leveys, double korkeus)
        : base(leveys, korkeus)
        {
            Mass = 1000;
            IsUpdated = true;
            IgnoresCollisionResponse = true;
        }
    }
}